/**
 * libFAD - Flash Animation Decode library
 * Copyright (C) 2005-2006 VGSystem Technologies, Inc.
 *
 * libFAD is the legal property of its developers, whose names are too numerous
 * to list here.  Please refer to the COPYRIGHT file distributed with this
 * source distribution.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: fad.frame.c,v 1.65 2006/03/22 01:44:59 wrxzzj Exp $
 */
#include "fad.types.h"
#include "fad.dl.h"
#include "fad.shape.h"
#include "fad.sound.h"
#include "fad.button.h"
#include "fad.fontext.h"
#include "fad.jpeg.h"
#include "fad.bitmap.h"
#include "fad.morph.h"
#include "fad.bits.h"
#include "fad.tags.h"
#include "fad.dynarray.h"
#include "fad.sprite.h"
#include "fad.actions.h"

typedef struct {
  const char* tag_string;
  s32_t (* tag_decode) (fad_frame_t* frame, fad_stream_t* s);
} fad_tag_t;

static fad_render_t default_render;

static fad_tag_t g_tag_array[] = {
  {"TAG_END", NULL },
  { "TAG_SHOWFRAME", NULL },
  { "TAG_DEFINESHAPE", style_shape_decode },
  { "TAG_FREECHARACTER", NULL },
  { "TAG_PLACEOBJECT", place_object_decode },
  { "TAG_REMOVEOBJECT", NULL },
  { "TAG_DEFINEBITS/JPEG", jpeg_bits_decode },
  { "TAG_DEFINEBUTTON", button_decode },
  { "TAG_JPEGTABLES", jpeg_tables_decode },
  { "TAG_SETBACKGROUNDCOLOR", NULL },
  { "TAG_DEFINEFONT", font_v1_decode },
  { "TAG_DEFINETEXT", static_text_decode },
#ifdef LIBFAD_AS2
  { "TAG_DOACTION", action_script_decode },
#else
  { "TAG_DOACTION", NULL },
#endif
  { "TAG_DEFINEFONTINFO", fontinfo_decode },
  { "TAG_DEFINESOUND", event_sound_decode },
  { "TAG_STARTSOUND", NULL },
  { "TAG_UNDEFINE(val=16)", NULL },
  { "TAG_DEFINEBUTTONSOUND", button_sound_decode },
  { "TAG_SOUNDSTREAMHEAD", stream_sound_head_decode },
  { "TAG_SOUNDSTREAMBLOCK", stream_sound_block_decode },
  { "TAG_DEFINEBITSLOSSLESS", bitmap_bits_decode },
  { "TAG_DEFINEBITSJPEG2", jpeg_bits_decode },
  { "TAG_DEFINESHAPE2", style_shape_decode },
  { "TAG_DEFINEBUTTONCXFORM", button_cxform_decode },
  { "TAG_PROTECT", NULL },
  { "TAG_UNDEFINE(val=25)", NULL },
  { "TAG_PLACEOBJECT2", place_object2_decode },
  { "TAG_UNDEFINE(val=27)", NULL },
  { "TAG_REMOVEOBJECT2", NULL },
  { "TAG_UNDEFINE(val=29)", NULL },
  { "TAG_UNDEFINE(val=30)", NULL },
  { "TAG_UNDEFINE(val=31)", NULL },
  { "TAG_DEFINESHAPE3", style_shape_decode },
  { "TAG_DEFINETEXT2", static_text_decode },
  { "TAG_DEFINEBUTTON2", button_decode },
  { "TAG_DEFINEBITSJPEG3", jpeg_bits_decode },
  { "TAG_DEFINEBITSLOSSLESS2", bitmap_bits_decode },
  { "TAG_DEFINEEDITTEXT", NULL },
  { "TAG_DEFINEMOVIE", NULL },
  { "TAG_DEFINESPRITE", sprite_mc_decode },
  { "TAG_NAMECHARACTER", NULL },
  { "TAG_SERIALNUMBER", NULL },
  { "TAG_GENERATORTEXT", NULL },
  { "TAG_FRAMELABEL", NULL },
  { "TAG_UNDEFINE(val=44)", NULL },
  { "TAG_SOUNDSTREAMHEAD2", stream_sound_head_decode },
  { "TAG_DEFINEMORPHSHAPE", morph_shape_decode },
  { "TAG_UNDEFINE(val=47)", NULL },
  { "TAG_DEFINEFONT2", font_v2_decode },
  { "TAG_TEMPLATECOMMAND", NULL  },
  { "TAG_GENERATOR3", NULL },
  { "TAG_UNDEFINE(val=51)", NULL },
  { "TAG_EXTERNALFONT", NULL },
  { "TAG_UNDEFINE(val=53)", NULL },
  { "TAG_UNDEFINE(val=54)", NULL },
  { "TAG_UNDEFINE(val=55)", NULL },
  { "TAG_EXPORTASSET", NULL },
  { "TAG_IMPORTASSETS", NULL },
  { "TAG_ENABLEDEBUGGER", NULL },
  { "TAG_DOINITACTION", NULL },
  { "TAG_DEFINEVIDEOSTREAM", NULL },
  { "TAG_VIDEOFRAME", NULL },
  { "TAG_DEFINEFONTINFO2", fontinfo_decode },
  { "TAG_MX4", NULL },
  { "TAG_ENABLEDEBUGGER2", NULL },
  { "TAG_SCRIPTLIMITS", NULL },
  { "TAG_SETTABINDEX", NULL }
};

static s32_t _fad_frame_proc_mbevt(fad_frame_t *frame, u8_t evt, s32_t x, s32_t y);

static s32_t _tag_decode_default_method (fad_frame_t* frame, fad_stream_t* s) {
  s32_t ret = 0;

  switch(s->tag_id) {
    case TAG_END:
      frame->sta = FAD_STA_FINISH;
      break;

    case TAG_SHOWFRAME:
      frame->output_frame++;
      frame->sta = FAD_STA_DORENDER;
      break;

    case TAG_STARTSOUND:
      {
        FAD_FRAME_SET_EVENTSOUND(frame->attr);
        frame->es_id = bits_get_u16(&s->bits);
        frame->es_si = bits_tell(&s->bits);
      }
      break;

    case TAG_REMOVEOBJECT:
      bits_get_u16(&s->bits);
    case TAG_REMOVEOBJECT2:
      {
        dl_node_t *node = NULL;
        u16_t depth = bits_get_u16(&s->bits);

        node = frame->dlhdr.rmv(&frame->dlhdr, depth);
        if(node)
          free(node);
        node = NULL;
      }
      break;

    case TAG_SETBACKGROUNDCOLOR:
      {
        u8_t* ptr = bits_tell(&s->bits);

        frame->r = *ptr; frame->g = *(ptr+1); frame->b = *(ptr+2);
        FAD_ERROR("frame background color(%d, %d, %d)\n", frame->r, frame->g, frame->b);
      }
      break;

    default:
      FAD_DEBUG("decode %s isn't implement now, ignore.\n", g_tag_array[s->tag_id].tag_string); 
      ret = -1;
      break;
  }

  return ret;
}



void fad_frame_init(fad_frame_t* frame) {
	frame_priv_t *priv = NULL;

  dl_header_init(&frame->dlhdr);

  frame->attr = 0;
	frame->nbtn = 0;
  frame->output_frame = 0;
  frame->render = &default_render;
  default_render_init(frame->render);
  frame->sta = FAD_STA_INIT;
  frame->proc_mbevt = _fad_frame_proc_mbevt;

	if(! frame->priv) {
		priv = frame->priv = calloc(1, sizeof(frame_priv_t));
	}

  mad_frame_init(&priv->mad_frame);
  mad_synth_init(&priv->mad_synth);
  mad_stream_init(&priv->mad_stream);
}

void fad_frame_finish(fad_frame_t* frame) {
	frame_priv_t *priv = NULL;
  dl_node_t* node = NULL;
  u16_t idx = 0, count = 0;

  dl_header_finish(&frame->dlhdr);

  default_render_term(frame->render);

	priv = (frame_priv_t *)frame->priv;
	if(priv) {
		mad_frame_finish(&priv->mad_frame);
		mad_synth_finish(&priv->mad_synth);
		mad_stream_finish(&priv->mad_stream);

		free(priv);
		frame->priv = NULL;
	}
}

s32_t fad_frame_seek(fad_frame_t *frame, fad_stream_t *s, s32_t idx) {
  s32_t offset = idx-frame->output_frame;

  if(offset > 0) {
    while(frame->output_frame < idx)
      fad_frame_decode(frame, s);
  } else if(offset < 0) {
    dl_header_finish(&frame->dlhdr);
    bits_buffer(&s->bits, s->bufbeg);

    frame->sta = FAD_STA_INIT;
    fad_frame_decode(frame, s);
    frame->sta = FAD_STA_SEEK;
    while(idx--)
      fad_frame_decode(frame, s);
  }

  return offset;
}

s32_t fad_frame_decode(fad_frame_t* frame, fad_stream_t* s) {
  u8_t* ptr = NULL;
  u16_t tag_id = 0;
  u32_t offset = 0;
  s32_t ret = 0, width = 0, height = 0;
  s32_t (* tag_decode_call)(fad_frame_t* frame, fad_stream_t* s);

  switch(frame->sta) {
    case FAD_STA_INIT:
      bits_get_rect(&s->bits, &frame->size);
      frame->rate = bits_get_u16(&s->bits);
      frame->total_frame = bits_get_u16(&s->bits);
      frame->sta = FAD_STA_DOTAG;
      FAD_ERROR("frame.total_frame = %d, frame.rate = %f, frame.size(%d, %d, %d, %d)\n\n",
             frame->total_frame, (double)frame->rate/0xff, frame->size.x0, frame->size.y0, frame->size.x1, frame->size.y1);
      break;

    case FAD_STA_SEEK:
    case FAD_STA_DOTAG:
      tag_id = bits_get_u16(&s->bits);
      s->tag_len = tag_id & 0x3F;
      tag_id = tag_id >> 6;
      if(s->tag_len == 0x3F)
        s->tag_len = bits_get_u32(&s->bits);

      if(tag_id > TAG_ID_MAX) {
        FAD_ERROR("tag id EXCEPTION, id = 0x%x, try skip.\n", tag_id);
        bits_seek_nbytes(&s->bits, s->tag_len);
        s->err = FAD_ERROR_TAG;
        ret = -1;
        break;
      } else {
        if(s->err == FAD_ERROR_TAG)
          s->err = FAD_ERROR_NONE;

        if(frame->sta == FAD_STA_SEEK) {
          if(tag_id!=TAG_PLACEOBJECT2 && tag_id!=TAG_PLACEOBJECT) {
            bits_seek_nbytes(&s->bits, s->tag_len);
            s->err = FAD_ERROR_NONE;
            ret = 0;
            break;
          }
        }
      }

      s->tag_id = tag_id;
      if(g_tag_array[s->tag_id].tag_decode)
        tag_decode_call = g_tag_array[s->tag_id].tag_decode;
      else
        tag_decode_call = _tag_decode_default_method;

      //bits_dump(&s->bits, 4);
      ptr = bits_tell(&s->bits);
      /**call do tag function by tag_id*/
      //FAD_ERROR("+++++%s, len = %d, offset = %x, is_sprite_frame = %x+++++\n", g_tag_array[s->tag_id].tag_string, s->tag_len, ptr-s->bufbeg, FAD_FRAME_CHK_SPRITE(frame->attr));
      ret = tag_decode_call(frame, s);
      offset = bits_tell(&s->bits) - ptr;

      if(ret < 0 || s->tag_len != offset) {
        s->bits.bufptr = ptr;
        bits_seek_nbytes(&s->bits, s->tag_len);
      }
      //FAD_ERROR("-----%s, ret = %d-----\n\n", g_tag_array[s->tag_id].tag_string, ret);
      break;
  }

  if(s->err != FAD_ERROR_NONE) {
    /**print error message even without LIBFAD_DEBUG defined*/
    fprintf(stderr, "***stream.err = %s, frame.sta = %s***\n", fad_stream_errstr(s->err), fad_frame_stastr(frame->sta));
  }
  return ret;
}

u8_t fad_frame_render_graphics(fad_frame_t *frame, dl_node_t *parent) {
  u8_t nbits, beof=FAD_TRUE, i;
  u16_t idx = 0;
  bits_t bits;
  dl_node_t *node = NULL;
  s32_t ret = 0;
  cairo_matrix_t *mx = NULL;

  bits_init(&bits);
  frame->nclip = -1;
  if(parent)
    mx = &parent->mx;
	else
		frame->nbtn = 0;

  for(node = frame->dlhdr.hdr; node; node = node->next) {
    cairo_matrix_t tempMX = {0};
    fad_render_t *dr = NULL;

    if(node == NULL || node->fo == NULL || ((node->flag&0x08) && (node->cx.am+node->cx.aa==0)))
      continue;

#ifdef LIBFAD_DIRTY_AREA
    if(fad_rect_check_over(&frame->dirty_rect, &node->rt) == FAD_FALSE)
      continue;
#endif

    FAD_ERROR("GET displaylist node = %x, depth = %d, count = %d, mx = %x, node->next=%x\n", node, node->depth, frame->dlhdr.count, mx, node->next);
    FAD_ERROR("(*)object (%s) depth = %d, matrix = (%f, %f, %f, %f, %f, %f)\n", fad_object_type_to_string(node->fo->type), node->depth, node->mx.xx, node->mx.xy, node->mx.x0, node->mx.yy, node->mx.yx, node->mx.y0);
    dr = frame->render;

    if(frame->nclip>=0 && frame->clipstack[frame->nclip] && node->depth>frame->clipstack[frame->nclip]) {
      cairo_restore(dr->cr);
      frame->clipstack[frame->nclip] = 0;
      frame->nclip--;
    }

    if(parent) {
      if(parent->flag&0x40)
        node->flag |= 0x40;

      if(FAD_DL_NODE_CHK_MATRIX(node->attr)) {
        if(!FAD_DL_NODE_CHK_MERGE(node->attr)) {
          cairo_matrix_multiply(&tempMX, &node->mx, mx);
          node->mx = tempMX;
          FAD_DL_NODE_SET_MERGE(node->attr);
        }
      } else {
        node->mx = *mx;
      }
    } else if(node->fo->type == FO_TYPE_BUTTON) {
			frame->nbtn++;
		}

    if(node->fo->type != FO_TYPE_SPRITE) {
      cairo_save(dr->cr);
      cairo_set_matrix(dr->cr, &node->mx);
    }
    if(node->fo->do_render) {
      if(node->fo->do_render(node->fo, dr, node) == FAD_FALSE)
        beof = FAD_FALSE;
    }
    if(node->fo->type != FO_TYPE_SPRITE)
      cairo_restore(dr->cr);

    FAD_ERROR("render object (%s,0x%x) done, depth = %d, nframe = %d, beof = %d.\n", fad_object_type_to_string(node->fo->type), node, node->depth, frame->output_frame, beof);

    if(node->ctx == NULL) continue;
    bits_buffer(&bits, node->ctx);

    if(node->flag & 0x20) {
      while(bits_get_u8(&bits) != 0);
    }

    if(node->flag & 0x40) {
      if(frame->nclip < FAD_CLIPSTACK_MAX-1) {
        u16_t clip = bits_get_u16(&bits);

        if((frame->nclip==-1) || (frame->nclip>=0 && clip<frame->clipstack[frame->nclip])) {
          frame->nclip++;
          frame->clipstack[frame->nclip] = clip;
          cairo_save(dr->cr);
          cairo_clip(dr->cr);
          cairo_new_path(dr->cr);
        }
      }
    }
  } /**for(....)*/

  for(i=0; i<=frame->nclip && frame->clipstack[i]; i++) {
    FAD_ERROR("CLEAR DIRTY CLIP STACK!!!\n");
    cairo_restore(frame->render->cr);
    frame->clipstack[i] = 0;
  }

  bits_finish(&bits);

  if(frame->sta != FAD_STA_FINISH && frame->sta != FAD_STA_UPDATEBTN && frame->sta != FAD_STA_END)
    frame->sta = FAD_STA_DOTAG;

  return beof;
}

void fad_frame_render_movie(fad_frame_t* frame) {

  if(fad_frame_render_graphics(frame, NULL) && frame->output_frame>=frame->total_frame) {
    fprintf(stderr, "*****THE END*****\n");
    frame->sta = FAD_STA_END;
  }
}

void fad_frame_render_audio(fad_frame_t *frame, fad_stream_t *s) {
  event_sound_t *es = NULL;
	frame_priv_t *priv = NULL;

  if(FAD_FRAME_CHK_EVENTSOUND(frame->attr)) {
    es = (event_sound_t *)s->dict->get(s->dict, frame->es_id);

    if(default_render_open_snddev(frame->render, es->attr) < 0)
      FAD_ERROR("event.sound: open /dev/dsp failure.\n");

    switch(es->attr&0xf0) {
      case 0x10: /**ADPCM audio decode*/
        fad_sound_render_adpcm(frame->render, es->data, es->attr, es->nsize/*4096<<(es->attr&0x01)*/);
        break;

      case 0x20: /**MPEG3 audio decode*/
        fad_sound_render_mpeg3(frame->render, es->data, es->attr, es->nsize, NULL, NULL, NULL);
        break;

      case 0x00:
      case 0x30: /**RAW audio decode*/
        FAD_ERROR("render raw sound data in event sound.\n");
        fad_sound_render_raw(frame->render, es->data, es->attr, es->nsize);
        break;

      case 0x60: /**Nellymoser audio decode*/
        FAD_ERROR("Nellymoser audio render is N/A.\n");
        break;
    }
  }

  if(FAD_FRAME_CHK_STREAMSOUND(frame->attr)) {
		priv = (frame_priv_t *)frame->priv;

    switch(frame->ss_attr&0xf0) {
      case 0x10: /**ADPCM audio decode*/
        fad_sound_render_adpcm(frame->render, frame->ss_buff, frame->ss_attr, frame->ss_nsize);
        break;

      case 0x20: /**MPEG3 audio decode*/
        fad_sound_render_mpeg3(frame->render, frame->ss_buff, frame->ss_attr, frame->ss_nsize, &priv->mad_frame, &priv->mad_synth, &priv->mad_stream);
        break;

      case 0x00:
      case 0x30: /**RAW audio decode*/
        FAD_ERROR("render raw sound data in stream sound.\n");
        fad_sound_render_raw(frame->render, frame->ss_buff, frame->ss_attr, frame->ss_nsize);
        break;

      case 0x60: /**Nellymoser audio decode*/
        FAD_ERROR("Nellymoser audio render is N/A.\n");
        break;
    }
  }

  FAD_FRAME_CLR_EVENTSOUND(frame->attr);
  FAD_FRAME_CLR_STREAMSOUND(frame->attr);
}

void fad_frame_render_actionscript(fad_frame_t* frame, fad_stream_t *stream) {
#ifdef LIBFAD_AS2
  fad_actions_context_t fac;

  if(frame->asr) {
    fad_actions_context_init(&fac, frame, stream);
    action_script_exec(&fac, frame->asr);
    fad_actions_context_finish(&fac);
  }
#endif
}

void fad_frame_set_update_callback(fad_frame_t *frame, void (* update)(void *, s32_t, s32_t, s32_t ,s32_t), void *surface) {
  if(frame) {
    frame->update = update;
    frame->priv_surface = surface;
  }
}

static s32_t _fad_frame_proc_mbevt(fad_frame_t *frame, u8_t evt, s32_t x, s32_t y) {
  dl_node_t *node = NULL;
	s32_t ret = 0, count = 0;

	x *= 20; y *= 20;
  for(node = frame->dlhdr.hdr; node && count<frame->nbtn; node = node->next) {
    if(node == NULL || node->fo == NULL || node->fo->type != FO_TYPE_BUTTON)
      continue;

		ret = node->fo->proc_mbevt(node->fo, evt, x, y);
		count++;

    if(ret>0 && (frame->sta==FAD_STA_END || frame->sta==FAD_STA_STOP))
      frame->sta = FAD_STA_UPDATEBTN;
//    printf("evt = %x, x = %d, y = %d, ret = %x\n", evt, x, y, ret);
  }

	return FAD_TRUE;
}

const char * fad_tag_id_to_string(u16_t tag_id) {
  if(tag_id < TAG_ID_MAX)
    return g_tag_array[tag_id].tag_string;

  return "TAG_UNDEFINE(val=???)";
}
